/**
 * @file core.c Galago Core message processor
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include "galagod.h"
#include "filter.h"
#include "service-list.h"
#include "person-list.h"
#include "utils.h"

static DBusHandlerResult
coco_add_service(DBusConnection *conn, DBusMessage *message, GalagodCoCo *coco)
{
	GalagoService *service;
	DBusMessageIter iter;

	dbus_message_iter_init(message, &iter);
	service = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_SERVICE);

	galago_object_set_data(service, "coco", coco);

	galagod_services_add(service);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
coco_remove_service(DBusConnection *conn, DBusMessage *message,
					GalagodCoCo *coco)
{
	GalagoService *service;
	DBusMessageIter iter;

	dbus_message_iter_init(message, &iter);
	service = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_SERVICE);

	galagod_services_remove(service);

	galago_object_unref(service);

	galagod_dbus_send_empty_reply(conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
coco_clear_services(DBusConnection *conn, DBusMessage *message,
					GalagodCoCo *coco)
{
	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

static DBusHandlerResult
coco_add_person(DBusConnection *conn, DBusMessage *message, GalagodCoCo *coco)
{
	GalagoPerson *person;
	DBusMessageIter iter;

	dbus_message_iter_init(message, &iter);
	person = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_PERSON);

	galago_object_set_data(person, "coco", coco);

	galago_log_misc("Adding person with ID %s (%p)\n",
					galago_person_get_id(person), person);
	galagod_people_add(person);

	galagod_dbus_send_empty_reply(conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
coco_remove_person(DBusConnection *conn, DBusMessage *message,
				   GalagodCoCo *coco)
{
	GalagoPerson *person;
	DBusMessageIter iter;

	dbus_message_iter_init(message, &iter);
	person = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_PERSON);

	galagod_people_remove(person);

	galago_object_unref(person);

	galagod_dbus_send_empty_reply(conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
coco_clear_people(DBusConnection *conn, DBusMessage *message,
				  GalagodCoCo *coco)
{
	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
coco_path_message_func(DBusConnection *conn, DBusMessage *message,
					   void *user_data)
{
	DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	GalagodCoCo *coco = (GalagodCoCo *)user_data;

	if (dbus_message_get_type(message) == DBUS_MESSAGE_TYPE_SIGNAL)
		return result;

	galago_log_misc("%s: %s\n",
					dbus_message_get_interface(message),
					dbus_message_get_member(message));

	galago_context_push(galagod_coco_get_context(coco));

	if (dbus_message_get_path(message) == NULL)
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;

	if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
									"AddService"))
	{
		result = coco_add_service(conn, message, coco);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "RemoveService"))
	{
		result = coco_remove_service(conn, message, coco);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "ClearServices"))
	{
		result = coco_clear_services(conn, message, coco);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "AddPerson"))
	{
		result = coco_add_person(conn, message, coco);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "RemovePerson"))
	{
		result = coco_remove_person(conn, message, coco);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "ClearPeople"))
	{
		result = coco_clear_people(conn, message, coco);
	}
	else
	{
		galago_log_warning("Unhandled message!\n");
		galago_log_misc("\tobj_path = '%s'\n",
						dbus_message_get_path(message));
		galago_log_misc("\tinterface = '%s'\n",
						dbus_message_get_interface(message));
		galago_log_misc("\tmethod = '%s'\n",
						dbus_message_get_member(message));
	}

	galago_context_pop();

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

static DBusObjectPathVTable coco_vtable =
{
	NULL,
	coco_path_message_func,
	NULL,
};

static DBusHandlerResult
core_register(DBusConnection *conn, DBusMessage *message, void *user_data)
{
	DBusConnection *dbus_conn;
	DBusMessage *reply;
	DBusMessageIter iter;
	GalagodCoCo *coco;
	gboolean feed;
	char *name;

	dbus_conn = galago_core_get_dbus_conn();

	dbus_message_iter_init(message, &iter);
	name = dbus_message_iter_get_string(&iter);
	dbus_message_iter_next(&iter);

	feed = dbus_message_iter_get_boolean(&iter);
	dbus_message_iter_next(&iter);

	coco = galagod_coco_new(name, feed, dbus_message_get_sender(message));

	if (coco == NULL)
	{
		reply = dbus_message_new_error(message, "RegistrationFailed",
									   "Registration failed.");
	}
	else
	{
		const char *obj_path;

		obj_path = galago_object_get_dbus_path(coco);

		galagod_dbus_register_object_path(obj_path, &coco_vtable, coco);

		reply = dbus_message_new_method_return(message);
		dbus_message_iter_init(reply, &iter);
		dbus_message_iter_append_string(&iter, galagod_coco_get_id(coco));
		dbus_message_iter_append_string(&iter, obj_path);
	}

	dbus_free(name);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
core_unregister(DBusConnection *conn, DBusMessage *message, void *user_data)
{
	DBusMessageIter iter;
	GalagodCoCo *coco;
	char *obj_path;

	dbus_message_iter_init(message, &iter);
	obj_path = dbus_message_iter_get_string(&iter);

	coco = galagod_cocos_get_with_obj_path(obj_path);

	if (coco == NULL)
	{
		galago_log_error("Attempted to unregister a non-existant CoCo "
						 "with path %s\n",
						 obj_path);

		dbus_free(obj_path);

		return DBUS_HANDLER_RESULT_HANDLED;
	}

	dbus_free(obj_path);

	galago_object_unref(coco);

	galagod_dbus_send_empty_reply(conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
core_get_service(DBusConnection *dbus_conn, DBusMessage *message,
				 void *user_data)
{
	DBusMessage *reply;
	DBusMessageIter iter;
	GalagodMetaService *meta_service;
	GalagoService *service = NULL;
	char *id;

	dbus_message_iter_init(message, &iter);
	id = dbus_message_iter_get_string(&iter);
	dbus_message_iter_next(&iter);

	meta_service = galagod_services_get_meta_service(id);

	if (meta_service != NULL)
		service = galagod_meta_service_get_service(meta_service);

	dbus_free(id);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);
	galago_dbus_message_iter_append_object(&iter, service);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
core_get_services(DBusConnection *dbus_conn, DBusMessage *message,
				  void *user_data)
{
	DBusMessage *reply;
	DBusMessageIter iter;
	const GalagoList *l;
	size_t service_count;
	GalagodMetaService *meta_service;
	GalagoService *service;

	service_count = galagod_services_get_meta_service_count();

	reply = dbus_message_new_method_return(message);

	dbus_message_iter_init(reply, &iter);
	dbus_message_iter_append_uint32(&iter, service_count);

	for (l = galagod_services_get_meta_services(); l != NULL; l = l->next)
	{
		meta_service = (GalagodMetaService *)l->data;
		service      = galagod_meta_service_get_service(meta_service);

		galago_dbus_message_iter_append_object(&iter, service);
	}

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
core_get_person(DBusConnection *dbus_conn, DBusMessage *message,
				void *user_data)
{
	DBusMessage *reply;
	DBusMessageIter iter;
	GalagodMetaPerson *meta_person;
	GalagoPerson *person = NULL;
	char *id;

	dbus_message_iter_init(message, &iter);
	id = dbus_message_iter_get_string(&iter);
	dbus_message_iter_next(&iter);

	meta_person = galagod_people_get_meta_person(id);

	if (meta_person != NULL)
		person = galagod_meta_person_get_person(meta_person);

	dbus_free(id);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);
	galago_dbus_message_iter_append_object(&iter, person);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
core_get_me(DBusConnection *dbus_conn, DBusMessage *message, void *user_data)
{
	DBusMessage *reply;
	DBusMessageIter iter;
	GalagodMetaPerson *meta_me;
	GalagoPerson *me = NULL;

	reply = dbus_message_new_method_return(message);

	meta_me = galagod_people_get_me();

	if (meta_me != NULL)
		me = galagod_meta_person_get_person(meta_me);

	dbus_message_iter_init(reply, &iter);
	galago_dbus_message_iter_append_object(&iter, me);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
core_get_people(DBusConnection *dbus_conn, DBusMessage *message,
				  void *user_data)
{
	DBusMessage *reply;
	DBusMessageIter iter;
	const GalagoList *l;
	size_t person_count;
	GalagodMetaPerson *meta_person;
	GalagoPerson *person;

	person_count = galagod_people_get_meta_person_count();

	reply = dbus_message_new_method_return(message);

	dbus_message_iter_init(reply, &iter);
	dbus_message_iter_append_uint32(&iter, person_count);

	for (l = galagod_people_get_meta_people(); l != NULL; l = l->next)
	{
		meta_person = (GalagodMetaPerson *)l->data;
		person      = galagod_meta_person_get_person(meta_person);

		galago_dbus_message_iter_append_object(&iter, person);
	}

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
core_shutdown(DBusConnection *dbus_conn, DBusMessage *message, void *user_data)
{
	galagod_dbus_send_empty_reply(dbus_conn, message);

	galagod_shutdown();

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
core_path_message_func(DBusConnection *conn, DBusMessage *message,
					   void *user_data)
{
	if (dbus_message_get_type(message) == DBUS_MESSAGE_TYPE_SIGNAL)
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;

	galago_log_misc("%s: %s\n",
					dbus_message_get_interface(message),
					dbus_message_get_member(message));

	if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
									"Register"))
	{
		return core_register(conn, message, user_data);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "Unregister"))
	{
		return core_unregister(conn, message, user_data);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "GetService"))
	{
		return core_get_service(conn, message, user_data);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "GetServices"))
	{
		return core_get_services(conn, message, user_data);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "GetMe"))
	{
		return core_get_me(conn, message, user_data);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "GetPerson"))
	{
		return core_get_person(conn, message, user_data);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "GetPeople"))
	{
		return core_get_people(conn, message, user_data);
	}
	else if (dbus_message_is_method_call(message, GALAGO_DBUS_CORE_INTERFACE,
										 "ShutDown"))
	{
		return core_shutdown(conn, message, user_data);
	}

	galago_log_warning("Unhandled message!\n");
	galago_log_misc("\tobj_path = '%s'\n",
					dbus_message_get_path(message));
	galago_log_misc("\tinterface = '%s'\n",
					dbus_message_get_interface(message));
	galago_log_misc("\tmethod = '%s'\n",
					dbus_message_get_member(message));

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

static DBusObjectPathVTable core_vtable =
{
	NULL,
	core_path_message_func,
	NULL,
};

static DBusHandlerResult
filter_func(DBusConnection *dbus_conn, DBusMessage *message, void *user_data)
{
	DBusMessageIter iter;

	if (dbus_message_is_signal(message,
							   DBUS_INTERFACE_ORG_FREEDESKTOP_DBUS,
							   "ServiceDeleted"))
	{
		char *dbus_service;

		dbus_message_iter_init(message, &iter);

		dbus_service = dbus_message_iter_get_string(&iter);

		if (strcmp(dbus_service, GALAGO_DBUS_SERVICE))
		{
			GalagodCoCo *coco;

			coco = galagod_cocos_get_with_service_id(dbus_service);

			if (coco != NULL)
				galago_object_unref(coco);
		}
	}

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

void
galagod_core_add_filter(void)
{
	DBusConnection *dbus_conn = galago_core_get_dbus_conn();

	galagod_dbus_register_object_path(GALAGO_DBUS_CORE_OBJECT, &core_vtable,
									  NULL);

	dbus_bus_add_match(dbus_conn,
					   "type='signal',"
					   "interface='" DBUS_INTERFACE_ORG_FREEDESKTOP_DBUS "',"
					   "sender='" DBUS_SERVICE_ORG_FREEDESKTOP_DBUS "'",
					   NULL);

	dbus_connection_add_filter(dbus_conn, filter_func, NULL, NULL);
}
